
#import <Cocoa/Cocoa.h>
#import <objc/runtime.h> //Make sure to import the runtime.h! 

void swizzle(void);
void swizzleBoolForKey(void);
void swizzleForKeyMethod(SEL selector);

BOOL fakeBoolForKey(id s, SEL cmd, NSString *defaultName);
id fakeAnythingForKey(id s, SEL cmd, NSString *defaultName);

//Static, global variable to hold the real implementation when we just swizzle boolForKey:
static IMP realMethodImplementation_boolForKey = NULL;

//Static, global dictionary to hold all of the real implementations if we choose to swizzle everything
static NSMutableDictionary *realMethodImplementations = NULL;


//The swizzled method to use when we want to just swizzle out boolForKey:
BOOL fakeBoolForKey(id s, SEL cmd, NSString *defaultName)
{

    NSString *result = @"No real method was found"; //The string that we'll log to the console
    BOOL r = NO; //The value we'll return to the caller
    
    //If we cached an real method, call it (remember, IMP is just a function pointer). 
    if(realMethodImplementation_boolForKey) { 
        
        //IMP is typed to return an id, but we can cast that to BOOL to get the right value
        r = (BOOL) realMethodImplementation_boolForKey(s, cmd, defaultName);

        //Pretty-print whatever we got back from the real function
        result = r ? @"YES" : @"NO"; 
    }
    
    NSLog(@"boolForKey: %@ and got answer: %@", defaultName, result);
    return r;
}

//The swizzled method to use when we want to swizzle out arbitrary {type}ForKey: 
id fakeAnythingForKey(id s, SEL cmd, NSString *defaultName)
{
    //Translate the selector to a string so we can log which method got called and 
    //so we can look up the real method implementation in our dictionary
    NSString *selectorString = NSStringFromSelector(cmd);
    
    NSLog(@"Called %@ for %@", selectorString, defaultName);
    
    //Look up the real implementation method in our dictionary using the selector string as key. 
    //IMP is a pointer, so we were able to store it in an NSValue
    NSValue *realImpValue = [realMethodImplementations valueForKey:selectorString];
    IMP realImp = NULL;
    
    //IMP is typed to return id
    id r = nil;
    if(realImp) {
        realImp = [realImpValue pointerValue]; //Get the function pointer value from the NSValue and call it
        r = realImp(s, cmd, defaultName);
    }
        
    //Since I don't know if r is really an object or a primative (were we objectForKey: or floatForKey:?),
    //I dont' want to log the output. If that were important, we could switch over the selector to format it
    //properly. Alternately, once we knew which method we were interested in, we could write a one-off swizzle 
    //like we did for boolForKey:
    return r;
}
               
//Inspiration and instruction for method swizzling from http://theocacao.com/document.page/327 
//and http://www.mikeash.com/pyblog/friday-qa-2010-01-29-method-replacement-for-fun-and-profit.html
void swizzleBoolForKey(void) 
{
    //We want to swizzle boolForKey: on NSUserDefaults, so get the basic metadata for those
    Class cls = [NSUserDefaults class];
    SEL selector = @selector(boolForKey:);
    
    //Convert the selector into a string so we can log which method we're swizzling
    NSString *selectorString = NSStringFromSelector(selector);
    
    //fakeBoolForKey(...) is the function we want to call instead of whatever Apple wrote. Turn it into 
    //an IMP (which is really a function pointer, so a simple cast will suffice)
    IMP newImplementation = (IMP) &fakeBoolForKey;
    
    //Lookup the method that Apple wrote
    Method realMethod = class_getInstanceMethod(cls, selector);
    
    //Save the function pointer for the method that Apple wrote so we can call it from the swizzled function
    realMethodImplementation_boolForKey = method_getImplementation(realMethod);
    
    //Try to replace Apple's IMP with ours. method_getTypeEncoding returns a char* encoding of the method's signature and 
    //parameters. Since we're swizzling out a method with identical parameters and signature, just get the type encoding from 
    //the real method. 
    if(!class_replaceMethod(cls, selector, newImplementation, method_getTypeEncoding(realMethod))) {
        NSLog(@"Could not replace %@", selectorString);
    } else {
        NSLog(@"Replaced %@!", selectorString);
    }
}

void swizzleForKeyMethod(SEL selector) 
{
    //We want to swizzle selector on NSUserDefaults, so get its class metadata
    Class cls = [NSUserDefaults class];
    
    //Convert the selector into a string so we can log which method we're swizzling and use it as a 
    //dictionary key for saving the real IMP to call from the swizzled method
    NSString *selectorString = NSStringFromSelector(selector);
    
    //We want to replace Apple's method with a fakeAnythingForKey. IMP is a function pointer so a simple cast
    //will suffice
    IMP newImplementation = (IMP) fakeAnythingForKey;
    
    //Lookup the method that Apple wrote
    Method realMethod = class_getInstanceMethod(cls, selector);
    
    //Get the function pointer for Apple's method and save it to an NSValue. Remember to use the alloc-init
    //pattern to avoid creating an autoreleased object with the convenience methods
    IMP realMethodImplementation = method_getImplementation(realMethod);
    
    NSValue *realMethodImplementationValue = [[NSValue alloc] initWithBytes:&realMethodImplementation objCType:@encode(IMP*)];

    //Store the implementation of the real method into a global dictionary with the selector string as the key
    [realMethodImplementations setValue:realMethodImplementationValue forKey:selectorString];
    
    //Try to replace Apple's IMP with ours. method_getTypeEncoding returns a char* encoding of the method's signature and 
    //parameters. Since we're swizzling out a method with identical parameters and signature, just get the type encoding from 
    //the real method. 
    if(!class_replaceMethod(cls, selector, newImplementation, method_getTypeEncoding(realMethod))) {
        NSLog(@"Could not replace %@", selectorString);
    } else {
        NSLog(@"Replaced %@!", selectorString);
    }
}

void swizzle(void)
{
    //Either swizzle out boolForKey: or use the generic swizzle to replace any or all of the 
    //{type}ForKey: methods
    
    //  swizzleBoolForKey();
    
    swizzleForKeyMethod(@selector(arrayForKey:));
    swizzleForKeyMethod(@selector(boolForKey:));
    swizzleForKeyMethod(@selector(dataForKey:));
    swizzleForKeyMethod(@selector(dictionaryForKey:));
    swizzleForKeyMethod(@selector(floatForKey:));
    swizzleForKeyMethod(@selector(integerForKey:));
    swizzleForKeyMethod(@selector(objectForKey:));
    swizzleForKeyMethod(@selector(stringArrayForKey:));
    swizzleForKeyMethod(@selector(stringForKey:));
    swizzleForKeyMethod(@selector(doubleForKey:));
    swizzleForKeyMethod(@selector(URLForKey:));
}

int main(int argc, char *argv[])
{
    //This needs to be done before calling swizzle() since swizzle() will use it.
    //REMEMBER: None of this code can do anything that would be autoreleased since NSApplicationMain
    //will eventually set up the autorelease pool. That means that this program leaks memory like a sieve! 
    realMethodImplementations = [[NSMutableDictionary alloc] init];
    
    swizzle();
    return NSApplicationMain(argc, (const char **)argv);
}
